@@ -1,9 +1,10 @@
+import collections
 import glob
 import os
 import platform
+import re
 import subprocess
 import sys
-from distutils.version import StrictVersion
 from itertools import chain
 
 import coverage
@@ -17,14 +18,17 @@ from process_tests import dump_on_error
 from process_tests import wait_for_strings
 from six import exec_
 
+import pytest_cov.plugin
+from pytest_cov import compat
+
 try:
     from StringIO import StringIO
 except ImportError:
     from io import StringIO
 
-import pytest_cov.plugin
+coverage, platform      # required for skipif mark on test_cov_min_from_coveragerc
 
-coverage, platform, StrictVersion  # required for skipif mark on test_cov_min_from_coveragerc
+max_worker_restart_0 = "--max-" + compat.worker + "-restart=0"
 
 SCRIPT = '''
 import sys, helper
@@ -245,7 +249,26 @@ def test_annotate_output_dir(testdir):
     assert result.ret == 0
 
 
-def test_html_output_dir(testdir, prop):
+def test_html(testdir):
+    script = testdir.makepyfile(SCRIPT)
+
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-report=html',
+                               script)
+
+    result.stdout.fnmatch_lines([
+        '*- coverage: platform *, python * -*',
+        'Coverage HTML written to dir htmlcov',
+        '*10 passed*',
+    ])
+    dest_dir = testdir.tmpdir.join('htmlcov')
+    assert dest_dir.check(dir=True)
+    assert dest_dir.join("index.html").check()
+    assert result.ret == 0
+
+
+def test_html_output_dir(testdir):
     script = testdir.makepyfile(SCRIPT)
 
     result = testdir.runpytest('-v',
@@ -264,6 +287,49 @@ def test_html_output_dir(testdir, prop):
     assert result.ret == 0
 
 
+def test_term_report_does_not_interact_with_html_output(testdir):
+    script = testdir.makepyfile(test_funcarg=SCRIPT_FUNCARG)
+
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-report=term-missing:skip-covered',
+                               '--cov-report=html:' + DEST_DIR,
+                               script)
+
+    result.stdout.fnmatch_lines([
+        '*- coverage: platform *, python * -*',
+        'Coverage HTML written to dir ' + DEST_DIR,
+        '*1 passed*',
+    ])
+    dest_dir = testdir.tmpdir.join(DEST_DIR)
+    assert dest_dir.check(dir=True)
+    assert sorted(dest_dir.visit("**/*.html")) == [dest_dir.join("index.html"), dest_dir.join("test_funcarg_py.html")]
+    assert dest_dir.join("index.html").check()
+    assert result.ret == 0
+
+
+def test_html_configured_output_dir(testdir):
+    script = testdir.makepyfile(SCRIPT)
+    testdir.tmpdir.join('.coveragerc').write("""
+[html]
+directory = somewhere
+""")
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-report=html',
+                               script)
+
+    result.stdout.fnmatch_lines([
+        '*- coverage: platform *, python * -*',
+        'Coverage HTML written to dir somewhere',
+        '*10 passed*',
+    ])
+    dest_dir = testdir.tmpdir.join('somewhere')
+    assert dest_dir.check(dir=True)
+    assert dest_dir.join("index.html").check()
+    assert result.ret == 0
+
+
 def test_xml_output_dir(testdir):
     script = testdir.makepyfile(SCRIPT)
 
@@ -330,7 +396,8 @@ def test_cov_min_50(testdir):
 
     result = testdir.runpytest('-v',
                                '--cov=%s' % script.dirpath(),
-                               '--cov-report=term-missing',
+                               '--cov-report=html',
+                               '--cov-report=xml',
                                '--cov-fail-under=50',
                                script)
 
@@ -340,6 +407,34 @@ def test_cov_min_50(testdir):
     ])
 
 
+def test_cov_min_float_value(testdir):
+    script = testdir.makepyfile(SCRIPT)
+
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-report=term-missing',
+                               '--cov-fail-under=88.88',
+                               script)
+    assert result.ret == 0
+    result.stdout.fnmatch_lines([
+        'Required test coverage of 88.88% reached. Total coverage: 88.89%'
+    ])
+
+
+def test_cov_min_float_value_not_reached(testdir):
+    script = testdir.makepyfile(SCRIPT)
+
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-report=term-missing',
+                               '--cov-fail-under=88.89',
+                               script)
+    assert result.ret == 1
+    result.stdout.fnmatch_lines([
+        'FAIL Required test coverage of 88.89% not reached. Total coverage: 88.89%'
+    ])
+
+
 def test_cov_min_no_report(testdir):
     script = testdir.makepyfile(SCRIPT)
 
@@ -375,7 +470,6 @@ def test_central_nonspecific(testdir, prop):
     assert result.ret == 0
 
 
-@pytest.mark.skipif('StrictVersion(coverage.__version__) <= StrictVersion("3.8")')
 def test_cov_min_from_coveragerc(testdir):
     script = testdir.makepyfile(SCRIPT)
     testdir.tmpdir.join('.coveragerc').write("""
@@ -557,6 +651,7 @@ def test_fail(p):
     result = testdir.runpytest('-v',
                                '--cov=%s' % script.dirpath(),
                                '--cov-fail-under=100',
+                               '--cov-report=html',
                                script)
 
     result.stdout.fnmatch_lines_random([
@@ -591,7 +686,7 @@ def test_foo(foo):
     ])
 
     for line in chain(result.stdout.lines, result.stderr.lines):
-        assert 'The following slaves failed to return coverage data' not in line
+        assert 'The following workers failed to return coverage data' not in line
         assert 'INTERNALERROR' not in line
     assert result.ret == 0
 
@@ -605,7 +700,7 @@ def test_dist_collocated(testdir, prop):
                                '--cov-report=term-missing',
                                '--dist=load',
                                '--tx=2*popen',
-                               '--max-slave-restart=0',
+                               max_worker_restart_0,
                                script, *prop.args)
 
     result.stdout.fnmatch_lines([
@@ -638,7 +733,7 @@ source =
                                '--tx=popen//chdir=%s' % dir2,
                                '--rsyncdir=%s' % script.basename,
                                '--rsyncdir=.coveragerc',
-                               '--max-slave-restart=0', '-s',
+                               max_worker_restart_0, '-s',
                                script, *prop.args)
 
     result.stdout.fnmatch_lines([
@@ -672,7 +767,7 @@ source =
                                '--tx=popen//chdir=%s' % dir2,
                                '--rsyncdir=%s' % script.basename,
                                '--rsyncdir=.coveragerc',
-                               '--max-slave-restart=0', '-s',
+                               max_worker_restart_0, '-s',
                                script, *prop.args)
 
     result.stdout.fnmatch_lines([
@@ -784,7 +879,7 @@ def test_dist_subprocess_collocated(testdir):
                                '--cov-report=term-missing',
                                '--dist=load',
                                '--tx=2*popen',
-                               '--max-slave-restart=0',
+                               max_worker_restart_0,
                                parent_script)
 
     result.stdout.fnmatch_lines([
@@ -819,7 +914,7 @@ source =
                                '--rsyncdir=%s' % child_script,
                                '--rsyncdir=%s' % parent_script,
                                '--rsyncdir=.coveragerc',
-                               '--max-slave-restart=0',
+                               max_worker_restart_0,
                                parent_script)
 
     result.stdout.fnmatch_lines([
@@ -859,6 +954,7 @@ def test_invalid_coverage_source(testdir):
 @pytest.mark.skipif("'dev' in pytest.__version__")
 @pytest.mark.skipif('sys.platform == "win32" and platform.python_implementation() == "PyPy"')
 def test_dist_missing_data(testdir):
+    """Test failure when using a worker without pytest-cov installed."""
     venv_path = os.path.join(str(testdir.tmpdir), 'venv')
     virtualenv.create_environment(venv_path)
     if sys.platform == 'win32':
@@ -880,17 +976,18 @@ def test_dist_missing_data(testdir):
     script = testdir.makepyfile(SCRIPT)
 
     result = testdir.runpytest('-v',
+                               '--assert=plain',
                                '--cov=%s' % script.dirpath(),
                                '--cov-report=term-missing',
                                '--dist=load',
                                '--tx=popen//python=%s' % exe,
-                               '--max-slave-restart=0',
+                               max_worker_restart_0,
                                script)
 
-    result.stdout.fnmatch_lines([
-        '*- coverage: failed slaves -*'
-    ])
     assert result.ret == 0
+    result.stdout.fnmatch_lines([
+        '*- coverage: failed workers -*'
+    ])
 
 
 def test_funcarg(testdir):
@@ -1006,6 +1103,7 @@ def test_run_target():
 
 
 @pytest.mark.skipif('sys.platform == "win32"', reason="multiprocessing support is broken on Windows")
+@pytest.mark.skipif('sys.version_info[0] > 2 and platform.python_implementation() == "PyPy"', reason="broken on PyPy3")
 def test_multiprocessing_pool_close(testdir):
     pytest.importorskip('multiprocessing.util')
 
@@ -1263,6 +1361,7 @@ if __name__ == "__main__":
 ''')
 
     result = testdir.runpytest('-vv',
+                               '--assert=plain',
                                '--cov=%s' % script.dirpath(),
                                '--cov-report=term-missing',
                                script)
@@ -1301,6 +1400,7 @@ if __name__ == "__main__":
 ''')
 
     result = testdir.runpytest('-vv',
+                               '--assert=plain',
                                '--cov=%s' % script.dirpath(),
                                '--cov-report=term-missing',
                                script)
@@ -1343,6 +1443,7 @@ if __name__ == "__main__":
     ''')
 
     result = testdir.runpytest('-vv',
+                               '--assert=plain',
                                '--cov=%s' % script.dirpath(),
                                '--cov-report=term-missing',
                                script)
@@ -1421,7 +1522,7 @@ def test_cover_conftest_dist(testdir):
                                '--cov-report=term-missing',
                                '--dist=load',
                                '--tx=2*popen',
-                               '--max-slave-restart=0',
+                               max_worker_restart_0,
                                script)
     assert result.ret == 0
     result.stdout.fnmatch_lines([CONF_RESULT])
@@ -1510,7 +1611,7 @@ def test_coveragerc_dist(testdir):
                                '--cov=%s' % script.dirpath(),
                                '--cov-report=term-missing',
                                '-n', '2',
-                               '--max-slave-restart=0',
+                               max_worker_restart_0,
                                script)
     assert result.ret == 0
     result.stdout.fnmatch_lines(
@@ -1536,7 +1637,6 @@ def test_basic():
 SKIP_COVERED_RESULT = '1 file skipped due to complete coverage.'
 
 
-@pytest.mark.skipif('StrictVersion(coverage.__version__) < StrictVersion("4.0")')
 @pytest.mark.parametrize('report_option', [
     'term-missing:skip-covered',
     'term:skip-covered'])
@@ -1551,7 +1651,6 @@ def test_skip_covered_cli(testdir, report_option):
     result.stdout.fnmatch_lines([SKIP_COVERED_RESULT])
 
 
-@pytest.mark.skipif('StrictVersion(coverage.__version__) < StrictVersion("4.0")')
 def test_skip_covered_coveragerc_config(testdir):
     testdir.makefile('', coveragerc=SKIP_COVERED_COVERAGERC)
     script = testdir.makepyfile(SKIP_COVERED_TEST)
@@ -1599,6 +1698,7 @@ def test_dist_boxed(testdir):
     script = testdir.makepyfile(SCRIPT_SIMPLE)
 
     result = testdir.runpytest('-v',
+                               '--assert=plain',
                                '--cov=%s' % script.dirpath(),
                                '--boxed',
                                script)
@@ -1612,6 +1712,8 @@ def test_dist_boxed(testdir):
 
 
 @pytest.mark.skipif('sys.platform == "win32"')
+@pytest.mark.skipif('sys.version_info[0] > 2 and platform.python_implementation() == "PyPy"',
+                    reason="strange optimization on PyPy3")
 def test_dist_bare_cov(testdir):
     script = testdir.makepyfile(SCRIPT_SIMPLE)
 
@@ -1658,7 +1760,11 @@ def test_disabled_output(testdir):
                                '--cov-report=',
                                script)
 
-    assert 'coverage' not in result.stdout.str()
+    stdout = result.stdout.str()
+    # We don't want the path to the executable to fail the test if we happen
+    # to put the project in a directory with "coverage" in it.
+    stdout = stdout.replace(sys.executable, "<SYS.EXECUTABLE>")
+    assert 'coverage' not in stdout
     assert result.ret == 0
 
 
@@ -1702,7 +1808,7 @@ data_file = %s
     result = testdir.runpytest('-v',
                                '--cov=%s' % script.dirpath(),
                                '-n', '1',
-                               '--max-slave-restart=0',
+                               max_worker_restart_0,
                                script)
     assert result.ret == 0
     assert glob.glob(str(testdir.tmpdir.join('some/special/place/coverage-data*')))
@@ -1789,6 +1895,7 @@ def test_pth_failure(monkeypatch):
 def test_double_cov(testdir):
     script = testdir.makepyfile(SCRIPT_SIMPLE)
     result = testdir.runpytest('-v',
+                               '--assert=plain',
                                '--cov', '--cov=%s' % script.dirpath(),
                                script)
 
@@ -1803,6 +1910,7 @@ def test_double_cov(testdir):
 def test_double_cov2(testdir):
     script = testdir.makepyfile(SCRIPT_SIMPLE)
     result = testdir.runpytest('-v',
+                               '--assert=plain',
                                '--cov', '--cov',
                                script)
 
@@ -1823,3 +1931,95 @@ def test_cov_and_no_cov(testdir):
                                script)
 
     assert result.ret == 0
+
+
+def find_labels(text, pattern):
+    all_labels = collections.defaultdict(list)
+    lines = text.splitlines()
+    for lineno, line in enumerate(lines, start=1):
+        labels = re.findall(pattern, line)
+        for label in labels:
+            all_labels[label].append(lineno)
+    return all_labels
+
+
+# The contexts and their labels in contextful.py
+EXPECTED_CONTEXTS = {
+    '': 'c0',
+    'test_contexts.py::test_01|run': 'r1',
+    'test_contexts.py::test_02|run': 'r2',
+    'test_contexts.py::OldStyleTests::test_03|setup': 's3',
+    'test_contexts.py::OldStyleTests::test_03|run': 'r3',
+    'test_contexts.py::OldStyleTests::test_04|run': 'r4',
+    'test_contexts.py::OldStyleTests::test_04|teardown': 't4',
+    'test_contexts.py::test_05|setup': 's5',
+    'test_contexts.py::test_05|run': 'r5',
+    'test_contexts.py::test_06|setup': 's6',
+    'test_contexts.py::test_06|run': 'r6',
+    'test_contexts.py::test_07|setup': 's7',
+    'test_contexts.py::test_07|run': 'r7',
+    'test_contexts.py::test_08|run': 'r8',
+    'test_contexts.py::test_09[1]|setup': 's9-1',
+    'test_contexts.py::test_09[1]|run': 'r9-1',
+    'test_contexts.py::test_09[2]|setup': 's9-2',
+    'test_contexts.py::test_09[2]|run': 'r9-2',
+    'test_contexts.py::test_09[3]|setup': 's9-3',
+    'test_contexts.py::test_09[3]|run': 'r9-3',
+    'test_contexts.py::test_10|run': 'r10',
+    'test_contexts.py::test_11[1-101]|run': 'r11-1',
+    'test_contexts.py::test_11[2-202]|run': 'r11-2',
+    'test_contexts.py::test_12[one]|run': 'r12-1',
+    'test_contexts.py::test_12[two]|run': 'r12-2',
+    'test_contexts.py::test_13[3-1]|run': 'r13-1',
+    'test_contexts.py::test_13[3-2]|run': 'r13-2',
+    'test_contexts.py::test_13[4-1]|run': 'r13-3',
+    'test_contexts.py::test_13[4-2]|run': 'r13-4',
+}
+
+
+@pytest.mark.skipif("coverage.version_info < (5, 0)")
+@xdist_params
+def test_contexts(testdir, opts):
+    with open(os.path.join(os.path.dirname(__file__), "contextful.py")) as f:
+        contextful_tests = f.read()
+    script = testdir.makepyfile(contextful_tests)
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-context=test',
+                               script,
+                               *opts.split()
+                               )
+    assert result.ret == 0
+    result.stdout.fnmatch_lines([
+        'test_contexts* 100%*',
+    ])
+
+    data = coverage.CoverageData(".coverage")
+    data.read()
+    assert data.measured_contexts() == set(EXPECTED_CONTEXTS)
+    measured = data.measured_files()
+    assert len(measured) == 1
+    test_context_path = list(measured)[0]
+    assert test_context_path.lower() == os.path.abspath("test_contexts.py").lower()
+
+    line_data = find_labels(contextful_tests, r"[crst]\d+(?:-\d+)?")
+    for context, label in EXPECTED_CONTEXTS.items():
+        if context == '':
+            continue
+        data.set_query_context(context)
+        actual = data.lines(test_context_path)
+        assert line_data[label] == actual, "Wrong lines for context {!r}".format(context)
+
+
+@pytest.mark.skipif("coverage.version_info >= (5, 0)")
+def test_contexts_not_supported(testdir):
+    script = testdir.makepyfile("a = 1")
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-context=test',
+                               script,
+                               )
+    result.stderr.fnmatch_lines([
+        '*argument --cov-context: Contexts are only supported with coverage.py >= 5.x',
+    ])
+    assert result.ret != 0
